#tab {
  overflow: hidden;
  border: 1px solid #ccc;
}


#tab button {
  background-color: inherit;
  float: left;
  border: none;
  outline: none;
  cursor: pointer;
  padding: 14px 16px;
  transition: 0.3s;
}

#tab button:hover {
  background-color: #ddd;
}

#tab button.active {
  background-color: #ccc;
}

.tabContent {
  display: none;
}