// NOTE: These definitions support NodeJS and TypeScript 3.2.
// This is required to enable typing assert in ts3.7 without causing errors
// Typically type modifications should be made in base.d.ts instead of here

/// <reference path="base.d.ts" />
/// <reference path="assert.d.ts" />
