DEFAULT_MIN_SIZE = 1
DEFAULT_DESIRED_CAPACITY = 1
DEFAULT_MAX_SIZE = 1

DEFAULT_INSTANCE_TYPE = "t2.micro"
DEFAULT_AMI_ID = "/aws/service/ami-amazon-linux-latest/amzn2-ami-hvm-x86_64-gp2"

BENTOSERVICE_PORT = 5000

AWS_EC2_IN_SERVICE_STATE = "InService"

TARGET_HEALTHY_STATUS = "healthy"
TARGET_UNHEALTHY_STATUS = "unhealthy"
TARGET_INITIAL_STATUS = "initial"
TARGET_UNUSED_STATUS = "unused"
TARGET_UNAVAILABLE_STATUS = "unavailable"
TARGET_DRAINIGN_STATUS = "draining"

TARGET_HEALTH_CHECK_INTERVAL = 5
TARGET_HEALTH_CHECK_PATH = "/healthz"
TARGET_HEALTH_CHECK_PORT = 5000
TARGET_HEALTH_CHECK_TIMEOUT_SECONDS = 3
TARGET_HEALTH_CHECK_THRESHOLD_COUNT = 2
