self.__precacheManifest = (self.__precacheManifest || []).concat([
  {
    "revision": "adc0e503f1993b199999b2d696d2d7ea",
    "url": "./index.html"
  },
  {
    "revision": "9f20598eaa054f5a7e87",
    "url": "./static/css/2.839a9555.chunk.css"
  },
  {
    "revision": "d9e1b8e2033fc08fa77b",
    "url": "./static/css/main.4c48667a.chunk.css"
  },
  {
    "revision": "9f20598eaa054f5a7e87",
    "url": "./static/js/2.d8f5929d.chunk.js"
  },
  {
    "revision": "0e9e4d1d98a593ee9ff6fc543843c9bd",
    "url": "./static/js/2.d8f5929d.chunk.js.LICENSE.txt"
  },
  {
    "revision": "d9e1b8e2033fc08fa77b",
    "url": "./static/js/main.f6224bc1.chunk.js"
  },
  {
    "revision": "2b66b08655ed491427c4",
    "url": "./static/js/runtime-main.d0e17700.js"
  },
  {
    "revision": "21d67caf31c9a92087a54e4754d15b97",
    "url": "./static/media/bentoml-logo.21d67caf.png"
  },
  {
    "revision": "05f1cdadfe476395f60e233b15c22155",
    "url": "./static/media/icons-16.05f1cdad.eot"
  },
  {
    "revision": "3c1c220e7a18286503fb431c7a7fe183",
    "url": "./static/media/icons-16.3c1c220e.woff"
  },
  {
    "revision": "3cde8748332d1de6b1ae1c2dc5850754",
    "url": "./static/media/icons-16.3cde8748.ttf"
  },
  {
    "revision": "0a5c76518a68c185baa2c6744456918c",
    "url": "./static/media/icons-20.0a5c7651.eot"
  },
  {
    "revision": "51ec31f302d0072808e1f83f85fea4cd",
    "url": "./static/media/icons-20.51ec31f3.ttf"
  },
  {
    "revision": "cef8cdbb9d0ba82e6e19fb0eeba2ac3d",
    "url": "./static/media/icons-20.cef8cdbb.woff"
  }
]);